"""Numba extension support for datetime.timedelta objects and their arrays.
"""
import datetime
import operator
from collections import namedtuple
import llvmlite.binding as ll
import numba
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_jitable, register_model, typeof_impl, unbox
from numba.parfors.array_analysis import ArrayAnalysis
import bodo
from bodo.hiframes.datetime_datetime_ext import datetime_datetime_type
from bodo.libs import hdatetime_ext
from bodo.utils.indexing import get_new_null_mask_bool_index, get_new_null_mask_int_index, get_new_null_mask_slice_index, setitem_slice_index_null_bits
from bodo.utils.typing import BodoError, get_overload_const_str, is_iterable_type, is_list_like_index_type, is_overload_constant_str
ll.add_symbol('box_datetime_timedelta_array', hdatetime_ext.
    box_datetime_timedelta_array)
ll.add_symbol('unbox_datetime_timedelta_array', hdatetime_ext.
    unbox_datetime_timedelta_array)


class NoInput:
    pass


_no_input = NoInput()


class NoInputType(types.Type):

    def __init__(self):
        super(NoInputType, self).__init__(name='NoInput')


register_model(NoInputType)(models.OpaqueModel)


@typeof_impl.register(NoInput)
def _typ_no_input(val, c):
    return NoInputType()


@lower_constant(NoInputType)
def constant_no_input(context, builder, ty, pyval):
    return context.get_dummy_value()


class PDTimeDeltaType(types.Type):

    def __init__(self):
        super(PDTimeDeltaType, self).__init__(name='PDTimeDeltaType()')


pd_timedelta_type = PDTimeDeltaType()
types.pd_timedelta_type = pd_timedelta_type


@typeof_impl.register(pd.Timedelta)
def typeof_pd_timedelta(val, c):
    return pd_timedelta_type


@register_model(PDTimeDeltaType)
class PDTimeDeltaModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        llzg__xsx = [('value', types.int64)]
        super(PDTimeDeltaModel, self).__init__(dmm, fe_type, llzg__xsx)


@box(PDTimeDeltaType)
def box_pd_timedelta(typ, val, c):
    nxrwd__gdq = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    psly__yvejy = c.pyapi.long_from_longlong(nxrwd__gdq.value)
    isha__nhp = c.pyapi.unserialize(c.pyapi.serialize_object(pd.Timedelta))
    res = c.pyapi.call_function_objargs(isha__nhp, (psly__yvejy,))
    c.pyapi.decref(psly__yvejy)
    c.pyapi.decref(isha__nhp)
    return res


@unbox(PDTimeDeltaType)
def unbox_pd_timedelta(typ, val, c):
    psly__yvejy = c.pyapi.object_getattr_string(val, 'value')
    pbawz__xems = c.pyapi.long_as_longlong(psly__yvejy)
    nxrwd__gdq = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    nxrwd__gdq.value = pbawz__xems
    c.pyapi.decref(psly__yvejy)
    ucckn__lctie = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(nxrwd__gdq._getvalue(), is_error=ucckn__lctie)


@lower_constant(PDTimeDeltaType)
def lower_constant_pd_timedelta(context, builder, ty, pyval):
    value = context.get_constant(types.int64, pyval.value)
    return lir.Constant.literal_struct([value])


@overload(pd.Timedelta, no_unliteral=True)
def pd_timedelta(value=_no_input, unit='ns', days=0, seconds=0,
    microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
    if value == _no_input:

        def impl_timedelta_kw(value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
            days += weeks * 7
            hours += days * 24
            minutes += 60 * hours
            seconds += 60 * minutes
            milliseconds += 1000 * seconds
            microseconds += 1000 * milliseconds
            dgsh__ndlan = 1000 * microseconds
            return init_pd_timedelta(dgsh__ndlan)
        return impl_timedelta_kw
    if value == bodo.string_type or is_overload_constant_str(value):

        def impl_str(value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
            with numba.objmode(res='pd_timedelta_type'):
                res = pd.Timedelta(value)
            return res
        return impl_str
    if value == pd_timedelta_type:
        return (lambda value=_no_input, unit='ns', days=0, seconds=0,
            microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0: value)
    if value == datetime_timedelta_type:

        def impl_timedelta_datetime(value=_no_input, unit='ns', days=0,
            seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0,
            weeks=0):
            days = value.days
            seconds = 60 * 60 * 24 * days + value.seconds
            microseconds = 1000 * 1000 * seconds + value.microseconds
            dgsh__ndlan = 1000 * microseconds
            return init_pd_timedelta(dgsh__ndlan)
        return impl_timedelta_datetime
    if not is_overload_constant_str(unit):
        raise BodoError('pd.to_timedelta(): unit should be a constant string')
    unit = pd._libs.tslibs.timedeltas.parse_timedelta_unit(
        get_overload_const_str(unit))
    latyd__obgm, elaap__klfp = pd._libs.tslibs.conversion.precision_from_unit(
        unit)

    def impl_timedelta(value=_no_input, unit='ns', days=0, seconds=0,
        microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
        return init_pd_timedelta(value * latyd__obgm)
    return impl_timedelta


@intrinsic
def init_pd_timedelta(typingctx, value):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        timedelta = cgutils.create_struct_proxy(typ)(context, builder)
        timedelta.value = args[0]
        return timedelta._getvalue()
    return PDTimeDeltaType()(value), codegen


make_attribute_wrapper(PDTimeDeltaType, 'value', '_value')


@overload_attribute(PDTimeDeltaType, 'value')
@overload_attribute(PDTimeDeltaType, 'delta')
def pd_timedelta_get_value(td):

    def impl(td):
        return td._value
    return impl


@overload_attribute(PDTimeDeltaType, 'days')
def pd_timedelta_get_days(td):

    def impl(td):
        return td._value // (1000 * 1000 * 1000 * 60 * 60 * 24)
    return impl


@overload_attribute(PDTimeDeltaType, 'seconds')
def pd_timedelta_get_seconds(td):

    def impl(td):
        return td._value // (1000 * 1000 * 1000) % (60 * 60 * 24)
    return impl


@overload_attribute(PDTimeDeltaType, 'microseconds')
def pd_timedelta_get_microseconds(td):

    def impl(td):
        return td._value // 1000 % 1000000
    return impl


@overload_attribute(PDTimeDeltaType, 'nanoseconds')
def pd_timedelta_get_nanoseconds(td):

    def impl(td):
        return td._value % 1000
    return impl


@register_jitable
def _to_hours_pd_td(td):
    return td._value // (1000 * 1000 * 1000 * 60 * 60) % 24


@register_jitable
def _to_minutes_pd_td(td):
    return td._value // (1000 * 1000 * 1000 * 60) % 60


@register_jitable
def _to_seconds_pd_td(td):
    return td._value // (1000 * 1000 * 1000) % 60


@register_jitable
def _to_milliseconds_pd_td(td):
    return td._value // (1000 * 1000) % 1000


@register_jitable
def _to_microseconds_pd_td(td):
    return td._value // 1000 % 1000


Components = namedtuple('Components', ['days', 'hours', 'minutes',
    'seconds', 'milliseconds', 'microseconds', 'nanoseconds'], defaults=[0,
    0, 0, 0, 0, 0, 0])


@overload_attribute(PDTimeDeltaType, 'components', no_unliteral=True)
def pd_timedelta_get_components(td):

    def impl(td):
        a = Components(td.days, _to_hours_pd_td(td), _to_minutes_pd_td(td),
            _to_seconds_pd_td(td), _to_milliseconds_pd_td(td),
            _to_microseconds_pd_td(td), td.nanoseconds)
        return a
    return impl


@overload_method(PDTimeDeltaType, '__hash__', no_unliteral=True)
def pd_td___hash__(td):

    def impl(td):
        return hash(td._value)
    return impl


@overload_method(PDTimeDeltaType, 'to_numpy', no_unliteral=True)
@overload_method(PDTimeDeltaType, 'to_timedelta64', no_unliteral=True)
def pd_td_to_numpy(td):
    from bodo.hiframes.pd_timestamp_ext import integer_to_timedelta64

    def impl(td):
        return integer_to_timedelta64(td.value)
    return impl


@overload_method(PDTimeDeltaType, 'to_pytimedelta', no_unliteral=True)
def pd_td_to_pytimedelta(td):

    def impl(td):
        return datetime.timedelta(microseconds=np.int64(td._value / 1000))
    return impl


@overload_method(PDTimeDeltaType, 'total_seconds', no_unliteral=True)
def pd_td_total_seconds(td):

    def impl(td):
        return td._value // 1000 / 10 ** 6
    return impl


def overload_add_operator_datetime_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            val = lhs.value + rhs.value
            return pd.Timedelta(val)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            qbpe__ldubf = (rhs.microseconds + (rhs.seconds + rhs.days * 60 *
                60 * 24) * 1000 * 1000) * 1000
            val = lhs.value + qbpe__ldubf
            return pd.Timedelta(val)
        return impl
    if lhs == datetime_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            ygjl__dluj = (lhs.microseconds + (lhs.seconds + lhs.days * 60 *
                60 * 24) * 1000 * 1000) * 1000
            val = ygjl__dluj + rhs.value
            return pd.Timedelta(val)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_datetime_type:
        from bodo.hiframes.pd_timestamp_ext import compute_pd_timestamp

        def impl(lhs, rhs):
            wtis__bkgu = rhs.toordinal()
            yvlj__rwo = rhs.second + rhs.minute * 60 + rhs.hour * 3600
            hbg__iqj = rhs.microsecond
            rpk__hxun = lhs.value // 1000
            cun__auuog = lhs.nanoseconds
            krykc__elkw = hbg__iqj + rpk__hxun
            amibq__grfn = 1000000 * (wtis__bkgu * 86400 + yvlj__rwo
                ) + krykc__elkw
            doxmp__vplv = cun__auuog
            return compute_pd_timestamp(amibq__grfn, doxmp__vplv)
        return impl
    if lhs == datetime_datetime_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + rhs.to_pytimedelta()
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs.days + rhs.days
            s = lhs.seconds + rhs.seconds
            us = lhs.microseconds + rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_datetime_type:

        def impl(lhs, rhs):
            frco__kkzlo = datetime.timedelta(rhs.toordinal(), hours=rhs.
                hour, minutes=rhs.minute, seconds=rhs.second, microseconds=
                rhs.microsecond)
            frco__kkzlo = frco__kkzlo + lhs
            oozyw__rnryv, vlxf__mhwaq = divmod(frco__kkzlo.seconds, 3600)
            xzj__etrr, eyje__qpf = divmod(vlxf__mhwaq, 60)
            if 0 < frco__kkzlo.days <= _MAXORDINAL:
                d = bodo.hiframes.datetime_date_ext.fromordinal_impl(
                    frco__kkzlo.days)
                return datetime.datetime(d.year, d.month, d.day,
                    oozyw__rnryv, xzj__etrr, eyje__qpf, frco__kkzlo.
                    microseconds)
            raise OverflowError('result out of range')
        return impl
    if lhs == datetime_datetime_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            frco__kkzlo = datetime.timedelta(lhs.toordinal(), hours=lhs.
                hour, minutes=lhs.minute, seconds=lhs.second, microseconds=
                lhs.microsecond)
            frco__kkzlo = frco__kkzlo + rhs
            oozyw__rnryv, vlxf__mhwaq = divmod(frco__kkzlo.seconds, 3600)
            xzj__etrr, eyje__qpf = divmod(vlxf__mhwaq, 60)
            if 0 < frco__kkzlo.days <= _MAXORDINAL:
                d = bodo.hiframes.datetime_date_ext.fromordinal_impl(
                    frco__kkzlo.days)
                return datetime.datetime(d.year, d.month, d.day,
                    oozyw__rnryv, xzj__etrr, eyje__qpf, frco__kkzlo.
                    microseconds)
            raise OverflowError('result out of range')
        return impl


def overload_sub_operator_datetime_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            lity__oonq = lhs.value - rhs.value
            return pd.Timedelta(lity__oonq)
        return impl
    if lhs == pd_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_datetime_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs.days - rhs.days
            s = lhs.seconds - rhs.seconds
            us = lhs.microseconds - rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl
    if lhs == datetime_datetime_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl
    if lhs == datetime_timedelta_array_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            szbek__ajjp = lhs
            numba.parfors.parfor.init_prange()
            n = len(szbek__ajjp)
            A = alloc_datetime_timedelta_array(n)
            for nrji__yfiw in numba.parfors.parfor.internal_prange(n):
                A[nrji__yfiw] = szbek__ajjp[nrji__yfiw] - rhs
            return A
        return impl


def overload_mul_operator_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value * rhs)
        return impl
    elif isinstance(lhs, types.Integer) and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return pd.Timedelta(rhs.value * lhs)
        return impl
    if lhs == datetime_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            d = lhs.days * rhs
            s = lhs.seconds * rhs
            us = lhs.microseconds * rhs
            return datetime.timedelta(d, s, us)
        return impl
    elif isinstance(lhs, types.Integer) and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            d = lhs * rhs.days
            s = lhs * rhs.seconds
            us = lhs * rhs.microseconds
            return datetime.timedelta(d, s, us)
        return impl


def overload_floordiv_operator_pd_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs.value // rhs.value
        return impl
    elif lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value // rhs)
        return impl


def overload_truediv_operator_pd_timedelta(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return lhs.value / rhs.value
        return impl
    elif lhs == pd_timedelta_type and isinstance(rhs, types.Integer):

        def impl(lhs, rhs):
            return pd.Timedelta(int(lhs.value / rhs))
        return impl


def overload_mod_operator_timedeltas(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            return pd.Timedelta(lhs.value % rhs.value)
        return impl
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            agvmi__zlot = _to_microseconds(lhs) % _to_microseconds(rhs)
            return datetime.timedelta(0, 0, agvmi__zlot)
        return impl


def pd_create_cmp_op_overload(op):

    def overload_pd_timedelta_cmp(lhs, rhs):
        if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

            def impl(lhs, rhs):
                return op(lhs.value, rhs.value)
            return impl
        if lhs == pd_timedelta_type and rhs == bodo.timedelta64ns:
            return lambda lhs, rhs: op(bodo.hiframes.pd_timestamp_ext.
                integer_to_timedelta64(lhs.value), rhs)
        if lhs == bodo.timedelta64ns and rhs == pd_timedelta_type:
            return lambda lhs, rhs: op(lhs, bodo.hiframes.pd_timestamp_ext.
                integer_to_timedelta64(rhs.value))
    return overload_pd_timedelta_cmp


@overload(operator.neg, no_unliteral=True)
def pd_timedelta_neg(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            return pd.Timedelta(-lhs.value)
        return impl


@overload(operator.pos, no_unliteral=True)
def pd_timedelta_pos(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            return lhs
        return impl


@overload(divmod, no_unliteral=True)
def pd_timedelta_divmod(lhs, rhs):
    if lhs == pd_timedelta_type and rhs == pd_timedelta_type:

        def impl(lhs, rhs):
            onp__qdlok, agvmi__zlot = divmod(lhs.value, rhs.value)
            return onp__qdlok, pd.Timedelta(agvmi__zlot)
        return impl


@overload(abs, no_unliteral=True)
def pd_timedelta_abs(lhs):
    if lhs == pd_timedelta_type:

        def impl(lhs):
            if lhs.value < 0:
                return -lhs
            else:
                return lhs
        return impl


class DatetimeTimeDeltaType(types.Type):

    def __init__(self):
        super(DatetimeTimeDeltaType, self).__init__(name=
            'DatetimeTimeDeltaType()')


datetime_timedelta_type = DatetimeTimeDeltaType()


@typeof_impl.register(datetime.timedelta)
def typeof_datetime_timedelta(val, c):
    return datetime_timedelta_type


@register_model(DatetimeTimeDeltaType)
class DatetimeTimeDeltaModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        llzg__xsx = [('days', types.int64), ('seconds', types.int64), (
            'microseconds', types.int64)]
        super(DatetimeTimeDeltaModel, self).__init__(dmm, fe_type, llzg__xsx)


@box(DatetimeTimeDeltaType)
def box_datetime_timedelta(typ, val, c):
    nxrwd__gdq = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    irb__lpr = c.pyapi.long_from_longlong(nxrwd__gdq.days)
    mrsz__thz = c.pyapi.long_from_longlong(nxrwd__gdq.seconds)
    mlh__oyck = c.pyapi.long_from_longlong(nxrwd__gdq.microseconds)
    isha__nhp = c.pyapi.unserialize(c.pyapi.serialize_object(datetime.
        timedelta))
    res = c.pyapi.call_function_objargs(isha__nhp, (irb__lpr, mrsz__thz,
        mlh__oyck))
    c.pyapi.decref(irb__lpr)
    c.pyapi.decref(mrsz__thz)
    c.pyapi.decref(mlh__oyck)
    c.pyapi.decref(isha__nhp)
    return res


@unbox(DatetimeTimeDeltaType)
def unbox_datetime_timedelta(typ, val, c):
    irb__lpr = c.pyapi.object_getattr_string(val, 'days')
    mrsz__thz = c.pyapi.object_getattr_string(val, 'seconds')
    mlh__oyck = c.pyapi.object_getattr_string(val, 'microseconds')
    cajfh__jbwty = c.pyapi.long_as_longlong(irb__lpr)
    ahf__heqzs = c.pyapi.long_as_longlong(mrsz__thz)
    izleu__cdvno = c.pyapi.long_as_longlong(mlh__oyck)
    nxrwd__gdq = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    nxrwd__gdq.days = cajfh__jbwty
    nxrwd__gdq.seconds = ahf__heqzs
    nxrwd__gdq.microseconds = izleu__cdvno
    c.pyapi.decref(irb__lpr)
    c.pyapi.decref(mrsz__thz)
    c.pyapi.decref(mlh__oyck)
    ucckn__lctie = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(nxrwd__gdq._getvalue(), is_error=ucckn__lctie)


@lower_constant(DatetimeTimeDeltaType)
def lower_constant_datetime_timedelta(context, builder, ty, pyval):
    days = context.get_constant(types.int64, pyval.days)
    seconds = context.get_constant(types.int64, pyval.seconds)
    microseconds = context.get_constant(types.int64, pyval.microseconds)
    return lir.Constant.literal_struct([days, seconds, microseconds])


@overload(datetime.timedelta, no_unliteral=True)
def datetime_timedelta(days=0, seconds=0, microseconds=0, milliseconds=0,
    minutes=0, hours=0, weeks=0):

    def impl_timedelta(days=0, seconds=0, microseconds=0, milliseconds=0,
        minutes=0, hours=0, weeks=0):
        d = s = us = 0
        days += weeks * 7
        seconds += minutes * 60 + hours * 3600
        microseconds += milliseconds * 1000
        d = days
        days, seconds = divmod(seconds, 24 * 3600)
        d += days
        s += int(seconds)
        seconds, us = divmod(microseconds, 1000000)
        days, seconds = divmod(seconds, 24 * 3600)
        d += days
        s += seconds
        return init_timedelta(d, s, us)
    return impl_timedelta


@intrinsic
def init_timedelta(typingctx, d, s, us):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        timedelta = cgutils.create_struct_proxy(typ)(context, builder)
        timedelta.days = args[0]
        timedelta.seconds = args[1]
        timedelta.microseconds = args[2]
        return timedelta._getvalue()
    return DatetimeTimeDeltaType()(d, s, us), codegen


make_attribute_wrapper(DatetimeTimeDeltaType, 'days', '_days')
make_attribute_wrapper(DatetimeTimeDeltaType, 'seconds', '_seconds')
make_attribute_wrapper(DatetimeTimeDeltaType, 'microseconds', '_microseconds')


@overload_attribute(DatetimeTimeDeltaType, 'days')
def timedelta_get_days(td):

    def impl(td):
        return td._days
    return impl


@overload_attribute(DatetimeTimeDeltaType, 'seconds')
def timedelta_get_seconds(td):

    def impl(td):
        return td._seconds
    return impl


@overload_attribute(DatetimeTimeDeltaType, 'microseconds')
def timedelta_get_microseconds(td):

    def impl(td):
        return td._microseconds
    return impl


@overload_method(DatetimeTimeDeltaType, 'total_seconds', no_unliteral=True)
def total_seconds(td):

    def impl(td):
        return ((td._days * 86400 + td._seconds) * 10 ** 6 + td._microseconds
            ) / 10 ** 6
    return impl


@overload_method(DatetimeTimeDeltaType, '__hash__', no_unliteral=True)
def __hash__(td):

    def impl(td):
        return hash((td._days, td._seconds, td._microseconds))
    return impl


@register_jitable
def _to_nanoseconds(td):
    return np.int64(((td._days * 86400 + td._seconds) * 1000000 + td.
        _microseconds) * 1000)


@register_jitable
def _to_microseconds(td):
    return (td._days * (24 * 3600) + td._seconds) * 1000000 + td._microseconds


@register_jitable
def _cmp(x, y):
    return 0 if x == y else 1 if x > y else -1


@register_jitable
def _getstate(td):
    return td._days, td._seconds, td._microseconds


@register_jitable
def _divide_and_round(a, b):
    onp__qdlok, agvmi__zlot = divmod(a, b)
    agvmi__zlot *= 2
    nojmq__sab = agvmi__zlot > b if b > 0 else agvmi__zlot < b
    if nojmq__sab or agvmi__zlot == b and onp__qdlok % 2 == 1:
        onp__qdlok += 1
    return onp__qdlok


_MAXORDINAL = 3652059


def overload_floordiv_operator_dt_timedelta(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return us // _to_microseconds(rhs)
        return impl
    elif lhs == datetime_timedelta_type and rhs == types.int64:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return datetime.timedelta(0, 0, us // rhs)
        return impl


def overload_truediv_operator_dt_timedelta(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return us / _to_microseconds(rhs)
        return impl
    elif lhs == datetime_timedelta_type and rhs == types.int64:

        def impl(lhs, rhs):
            us = _to_microseconds(lhs)
            return datetime.timedelta(0, 0, _divide_and_round(us, rhs))
        return impl


def create_cmp_op_overload(op):

    def overload_timedelta_cmp(lhs, rhs):
        if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

            def impl(lhs, rhs):
                lgs__gsur = _cmp(_getstate(lhs), _getstate(rhs))
                return op(lgs__gsur, 0)
            return impl
    return overload_timedelta_cmp


@overload(operator.neg, no_unliteral=True)
def timedelta_neg(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            return datetime.timedelta(-lhs.days, -lhs.seconds, -lhs.
                microseconds)
        return impl


@overload(operator.pos, no_unliteral=True)
def timedelta_pos(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            return lhs
        return impl


@overload(divmod, no_unliteral=True)
def timedelta_divmod(lhs, rhs):
    if lhs == datetime_timedelta_type and rhs == datetime_timedelta_type:

        def impl(lhs, rhs):
            onp__qdlok, agvmi__zlot = divmod(_to_microseconds(lhs),
                _to_microseconds(rhs))
            return onp__qdlok, datetime.timedelta(0, 0, agvmi__zlot)
        return impl


@overload(abs, no_unliteral=True)
def timedelta_abs(lhs):
    if lhs == datetime_timedelta_type:

        def impl(lhs):
            if lhs.days < 0:
                return -lhs
            else:
                return lhs
        return impl


@intrinsic
def cast_numpy_timedelta_to_int(typingctx, val=None):
    assert val in (types.NPTimedelta('ns'), types.int64)

    def codegen(context, builder, signature, args):
        return args[0]
    return types.int64(val), codegen


@overload(bool, no_unliteral=True)
def timedelta_to_bool(timedelta):
    if timedelta != datetime_timedelta_type:
        return
    wzxpp__rxxf = datetime.timedelta(0)

    def impl(timedelta):
        return timedelta != wzxpp__rxxf
    return impl


class DatetimeTimeDeltaArrayType(types.ArrayCompatible):

    def __init__(self):
        super(DatetimeTimeDeltaArrayType, self).__init__(name=
            'DatetimeTimeDeltaArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return datetime_timedelta_type

    def copy(self):
        return DatetimeTimeDeltaArrayType()


datetime_timedelta_array_type = DatetimeTimeDeltaArrayType()
types.datetime_timedelta_array_type = datetime_timedelta_array_type
days_data_type = types.Array(types.int64, 1, 'C')
seconds_data_type = types.Array(types.int64, 1, 'C')
microseconds_data_type = types.Array(types.int64, 1, 'C')
nulls_type = types.Array(types.uint8, 1, 'C')


@register_model(DatetimeTimeDeltaArrayType)
class DatetimeTimeDeltaArrayModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        llzg__xsx = [('days_data', days_data_type), ('seconds_data',
            seconds_data_type), ('microseconds_data',
            microseconds_data_type), ('null_bitmap', nulls_type)]
        models.StructModel.__init__(self, dmm, fe_type, llzg__xsx)


make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'days_data', '_days_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'seconds_data',
    '_seconds_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'microseconds_data',
    '_microseconds_data')
make_attribute_wrapper(DatetimeTimeDeltaArrayType, 'null_bitmap',
    '_null_bitmap')


@overload_method(DatetimeTimeDeltaArrayType, 'copy', no_unliteral=True)
def overload_datetime_timedelta_arr_copy(A):
    return (lambda A: bodo.hiframes.datetime_timedelta_ext.
        init_datetime_timedelta_array(A._days_data.copy(), A._seconds_data.
        copy(), A._microseconds_data.copy(), A._null_bitmap.copy()))


@unbox(DatetimeTimeDeltaArrayType)
def unbox_datetime_timedelta_array(typ, val, c):
    n = bodo.utils.utils.object_length(c, val)
    ncl__vpc = types.Array(types.intp, 1, 'C')
    vkzle__zbswy = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        ncl__vpc, [n])
    qzt__lodib = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        ncl__vpc, [n])
    zisg__uipd = bodo.utils.utils._empty_nd_impl(c.context, c.builder,
        ncl__vpc, [n])
    lzh__pjdr = c.builder.udiv(c.builder.add(n, lir.Constant(lir.IntType(64
        ), 7)), lir.Constant(lir.IntType(64), 8))
    pff__hil = bodo.utils.utils._empty_nd_impl(c.context, c.builder, types.
        Array(types.uint8, 1, 'C'), [lzh__pjdr])
    mrgen__hllvc = lir.FunctionType(lir.VoidType(), [lir.IntType(8).
        as_pointer(), lir.IntType(64), lir.IntType(64).as_pointer(), lir.
        IntType(64).as_pointer(), lir.IntType(64).as_pointer(), lir.IntType
        (8).as_pointer()])
    aztkj__gxv = cgutils.get_or_insert_function(c.builder.module,
        mrgen__hllvc, name='unbox_datetime_timedelta_array')
    c.builder.call(aztkj__gxv, [val, n, vkzle__zbswy.data, qzt__lodib.data,
        zisg__uipd.data, pff__hil.data])
    syab__guez = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    syab__guez.days_data = vkzle__zbswy._getvalue()
    syab__guez.seconds_data = qzt__lodib._getvalue()
    syab__guez.microseconds_data = zisg__uipd._getvalue()
    syab__guez.null_bitmap = pff__hil._getvalue()
    ucckn__lctie = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(syab__guez._getvalue(), is_error=ucckn__lctie)


@box(DatetimeTimeDeltaArrayType)
def box_datetime_timedelta_array(typ, val, c):
    szbek__ajjp = cgutils.create_struct_proxy(typ)(c.context, c.builder, val)
    vkzle__zbswy = c.context.make_array(types.Array(types.int64, 1, 'C'))(c
        .context, c.builder, szbek__ajjp.days_data)
    qzt__lodib = c.context.make_array(types.Array(types.int64, 1, 'C'))(c.
        context, c.builder, szbek__ajjp.seconds_data).data
    zisg__uipd = c.context.make_array(types.Array(types.int64, 1, 'C'))(c.
        context, c.builder, szbek__ajjp.microseconds_data).data
    srqtd__nreoh = c.context.make_array(types.Array(types.uint8, 1, 'C'))(c
        .context, c.builder, szbek__ajjp.null_bitmap).data
    n = c.builder.extract_value(vkzle__zbswy.shape, 0)
    mrgen__hllvc = lir.FunctionType(c.pyapi.pyobj, [lir.IntType(64), lir.
        IntType(64).as_pointer(), lir.IntType(64).as_pointer(), lir.IntType
        (64).as_pointer(), lir.IntType(8).as_pointer()])
    byci__zniuf = cgutils.get_or_insert_function(c.builder.module,
        mrgen__hllvc, name='box_datetime_timedelta_array')
    tlesx__zqo = c.builder.call(byci__zniuf, [n, vkzle__zbswy.data,
        qzt__lodib, zisg__uipd, srqtd__nreoh])
    c.context.nrt.decref(c.builder, typ, val)
    return tlesx__zqo


@intrinsic
def init_datetime_timedelta_array(typingctx, days_data, seconds_data,
    microseconds_data, nulls=None):
    assert days_data == types.Array(types.int64, 1, 'C')
    assert seconds_data == types.Array(types.int64, 1, 'C')
    assert microseconds_data == types.Array(types.int64, 1, 'C')
    assert nulls == types.Array(types.uint8, 1, 'C')

    def codegen(context, builder, signature, args):
        ime__wgo, bkqb__fnzwn, daci__bvzud, agpgi__kye = args
        wckk__niv = cgutils.create_struct_proxy(signature.return_type)(context,
            builder)
        wckk__niv.days_data = ime__wgo
        wckk__niv.seconds_data = bkqb__fnzwn
        wckk__niv.microseconds_data = daci__bvzud
        wckk__niv.null_bitmap = agpgi__kye
        context.nrt.incref(builder, signature.args[0], ime__wgo)
        context.nrt.incref(builder, signature.args[1], bkqb__fnzwn)
        context.nrt.incref(builder, signature.args[2], daci__bvzud)
        context.nrt.incref(builder, signature.args[3], agpgi__kye)
        return wckk__niv._getvalue()
    hgek__lcvu = datetime_timedelta_array_type(days_data, seconds_data,
        microseconds_data, nulls)
    return hgek__lcvu, codegen


@lower_constant(DatetimeTimeDeltaArrayType)
def lower_constant_datetime_timedelta_arr(context, builder, typ, pyval):
    n = len(pyval)
    vkzle__zbswy = np.empty(n, np.int64)
    qzt__lodib = np.empty(n, np.int64)
    zisg__uipd = np.empty(n, np.int64)
    xaoud__qjci = np.empty(n + 7 >> 3, np.uint8)
    for nrji__yfiw, s in enumerate(pyval):
        ffub__fkxbu = pd.isna(s)
        bodo.libs.int_arr_ext.set_bit_to_arr(xaoud__qjci, nrji__yfiw, int(
            not ffub__fkxbu))
        if not ffub__fkxbu:
            vkzle__zbswy[nrji__yfiw] = s.days
            qzt__lodib[nrji__yfiw] = s.seconds
            zisg__uipd[nrji__yfiw] = s.microseconds
    ebsxu__wfmm = context.get_constant_generic(builder, days_data_type,
        vkzle__zbswy)
    abud__bep = context.get_constant_generic(builder, seconds_data_type,
        qzt__lodib)
    ikiof__dgu = context.get_constant_generic(builder,
        microseconds_data_type, zisg__uipd)
    bqmco__gur = context.get_constant_generic(builder, nulls_type, xaoud__qjci)
    return lir.Constant.literal_struct([ebsxu__wfmm, abud__bep, ikiof__dgu,
        bqmco__gur])


@numba.njit(no_cpython_wrapper=True)
def alloc_datetime_timedelta_array(n):
    vkzle__zbswy = np.empty(n, dtype=np.int64)
    qzt__lodib = np.empty(n, dtype=np.int64)
    zisg__uipd = np.empty(n, dtype=np.int64)
    nulls = np.full(n + 7 >> 3, 255, np.uint8)
    return init_datetime_timedelta_array(vkzle__zbswy, qzt__lodib,
        zisg__uipd, nulls)


def alloc_datetime_timedelta_array_equiv(self, scope, equiv_set, loc, args, kws
    ):
    assert len(args) == 1 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


(ArrayAnalysis.
    _analyze_op_call_bodo_hiframes_datetime_timedelta_ext_alloc_datetime_timedelta_array
    ) = alloc_datetime_timedelta_array_equiv


@overload(operator.getitem, no_unliteral=True)
def dt_timedelta_arr_getitem(A, ind):
    if A != datetime_timedelta_array_type:
        return
    if isinstance(ind, types.Integer):

        def impl_int(A, ind):
            return datetime.timedelta(days=A._days_data[ind], seconds=A.
                _seconds_data[ind], microseconds=A._microseconds_data[ind])
        return impl_int
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:

        def impl_bool(A, ind):
            mwpqi__hemxp = bodo.utils.conversion.coerce_to_ndarray(ind)
            uuqq__iyzyc = A._null_bitmap
            vgcry__fzrus = A._days_data[mwpqi__hemxp]
            pnhx__abm = A._seconds_data[mwpqi__hemxp]
            ymwf__nltkg = A._microseconds_data[mwpqi__hemxp]
            n = len(vgcry__fzrus)
            qlb__jkqjs = get_new_null_mask_bool_index(uuqq__iyzyc, ind, n)
            return init_datetime_timedelta_array(vgcry__fzrus, pnhx__abm,
                ymwf__nltkg, qlb__jkqjs)
        return impl_bool
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):

        def impl(A, ind):
            mwpqi__hemxp = bodo.utils.conversion.coerce_to_ndarray(ind)
            uuqq__iyzyc = A._null_bitmap
            vgcry__fzrus = A._days_data[mwpqi__hemxp]
            pnhx__abm = A._seconds_data[mwpqi__hemxp]
            ymwf__nltkg = A._microseconds_data[mwpqi__hemxp]
            n = len(vgcry__fzrus)
            qlb__jkqjs = get_new_null_mask_int_index(uuqq__iyzyc,
                mwpqi__hemxp, n)
            return init_datetime_timedelta_array(vgcry__fzrus, pnhx__abm,
                ymwf__nltkg, qlb__jkqjs)
        return impl
    if isinstance(ind, types.SliceType):

        def impl_slice(A, ind):
            n = len(A._days_data)
            uuqq__iyzyc = A._null_bitmap
            vgcry__fzrus = np.ascontiguousarray(A._days_data[ind])
            pnhx__abm = np.ascontiguousarray(A._seconds_data[ind])
            ymwf__nltkg = np.ascontiguousarray(A._microseconds_data[ind])
            qlb__jkqjs = get_new_null_mask_slice_index(uuqq__iyzyc, ind, n)
            return init_datetime_timedelta_array(vgcry__fzrus, pnhx__abm,
                ymwf__nltkg, qlb__jkqjs)
        return impl_slice
    raise BodoError(
        f'getitem for DatetimeTimedeltaArray with indexing type {ind} not supported.'
        )


@overload(operator.setitem, no_unliteral=True)
def dt_timedelta_arr_setitem(A, ind, val):
    if A != datetime_timedelta_array_type:
        return
    if val == types.none or isinstance(val, types.optional):
        return
    nbruc__dmsj = (
        f"setitem for DatetimeTimedeltaArray with indexing type {ind} received an incorrect 'value' type {val}."
        )
    if isinstance(ind, types.Integer):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl(A, ind, val):
                A._days_data[ind] = val._days
                A._seconds_data[ind] = val._seconds
                A._microseconds_data[ind] = val._microseconds
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, ind, 1)
            return impl
        else:
            raise BodoError(nbruc__dmsj)
    if not (is_iterable_type(val) and val.dtype == bodo.
        datetime_timedelta_type or types.unliteral(val) ==
        datetime_timedelta_type):
        raise BodoError(nbruc__dmsj)
    if is_list_like_index_type(ind) and isinstance(ind.dtype, types.Integer):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_arr_ind_scalar(A, ind, val):
                n = len(A)
                for nrji__yfiw in range(n):
                    A._days_data[ind[nrji__yfiw]] = val._days
                    A._seconds_data[ind[nrji__yfiw]] = val._seconds
                    A._microseconds_data[ind[nrji__yfiw]] = val._microseconds
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ind[nrji__yfiw], 1)
            return impl_arr_ind_scalar
        else:

            def impl_arr_ind(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(val._days_data)
                for nrji__yfiw in range(n):
                    A._days_data[ind[nrji__yfiw]] = val._days_data[nrji__yfiw]
                    A._seconds_data[ind[nrji__yfiw]] = val._seconds_data[
                        nrji__yfiw]
                    A._microseconds_data[ind[nrji__yfiw]
                        ] = val._microseconds_data[nrji__yfiw]
                    sjai__ottzg = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, nrji__yfiw)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ind[nrji__yfiw], sjai__ottzg)
            return impl_arr_ind
    if is_list_like_index_type(ind) and ind.dtype == types.bool_:
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_bool_ind_mask_scalar(A, ind, val):
                n = len(ind)
                for nrji__yfiw in range(n):
                    if not bodo.libs.array_kernels.isna(ind, nrji__yfiw
                        ) and ind[nrji__yfiw]:
                        A._days_data[nrji__yfiw] = val._days
                        A._seconds_data[nrji__yfiw] = val._seconds
                        A._microseconds_data[nrji__yfiw] = val._microseconds
                        bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                            nrji__yfiw, 1)
            return impl_bool_ind_mask_scalar
        else:

            def impl_bool_ind_mask(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(ind)
                rnjt__hhojk = 0
                for nrji__yfiw in range(n):
                    if not bodo.libs.array_kernels.isna(ind, nrji__yfiw
                        ) and ind[nrji__yfiw]:
                        A._days_data[nrji__yfiw] = val._days_data[rnjt__hhojk]
                        A._seconds_data[nrji__yfiw] = val._seconds_data[
                            rnjt__hhojk]
                        A._microseconds_data[nrji__yfiw
                            ] = val._microseconds_data[rnjt__hhojk]
                        sjai__ottzg = bodo.libs.int_arr_ext.get_bit_bitmap_arr(
                            val._null_bitmap, rnjt__hhojk)
                        bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                            nrji__yfiw, sjai__ottzg)
                        rnjt__hhojk += 1
            return impl_bool_ind_mask
    if isinstance(ind, types.SliceType):
        if types.unliteral(val) == datetime_timedelta_type:

            def impl_slice_scalar(A, ind, val):
                xux__bpkk = numba.cpython.unicode._normalize_slice(ind, len(A))
                for nrji__yfiw in range(xux__bpkk.start, xux__bpkk.stop,
                    xux__bpkk.step):
                    A._days_data[nrji__yfiw] = val._days
                    A._seconds_data[nrji__yfiw] = val._seconds
                    A._microseconds_data[nrji__yfiw] = val._microseconds
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        nrji__yfiw, 1)
            return impl_slice_scalar
        else:

            def impl_slice_mask(A, ind, val):
                val = bodo.utils.conversion.coerce_to_array(val,
                    use_nullable_array=True)
                n = len(A._days_data)
                A._days_data[ind] = val._days_data
                A._seconds_data[ind] = val._seconds_data
                A._microseconds_data[ind] = val._microseconds_data
                obo__wpyqa = val._null_bitmap.copy()
                setitem_slice_index_null_bits(A._null_bitmap, obo__wpyqa,
                    ind, n)
            return impl_slice_mask
    raise BodoError(
        f'setitem for DatetimeTimedeltaArray with indexing type {ind} not supported.'
        )


@overload(len, no_unliteral=True)
def overload_len_datetime_timedelta_arr(A):
    if A == datetime_timedelta_array_type:
        return lambda A: len(A._days_data)


@overload_attribute(DatetimeTimeDeltaArrayType, 'shape')
def overload_datetime_timedelta_arr_shape(A):
    return lambda A: (len(A._days_data),)


@overload_attribute(DatetimeTimeDeltaArrayType, 'nbytes')
def timedelta_arr_nbytes_overload(A):
    return (lambda A: A._days_data.nbytes + A._seconds_data.nbytes + A.
        _microseconds_data.nbytes + A._null_bitmap.nbytes)


def overload_datetime_timedelta_arr_sub(arg1, arg2):
    if (arg1 == datetime_timedelta_array_type and arg2 ==
        datetime_timedelta_type):

        def impl(arg1, arg2):
            szbek__ajjp = arg1
            numba.parfors.parfor.init_prange()
            n = len(szbek__ajjp)
            A = alloc_datetime_timedelta_array(n)
            for nrji__yfiw in numba.parfors.parfor.internal_prange(n):
                A[nrji__yfiw] = szbek__ajjp[nrji__yfiw] - arg2
            return A
        return impl


def create_cmp_op_overload_arr(op):

    def overload_date_arr_cmp(lhs, rhs):
        if op == operator.ne:
            pbrh__hmlj = True
        else:
            pbrh__hmlj = False
        if (lhs == datetime_timedelta_array_type and rhs ==
            datetime_timedelta_array_type):

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(lhs)
                wfr__xst = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for nrji__yfiw in numba.parfors.parfor.internal_prange(n):
                    hblra__png = bodo.libs.array_kernels.isna(lhs, nrji__yfiw)
                    aqkel__vfsx = bodo.libs.array_kernels.isna(rhs, nrji__yfiw)
                    if hblra__png or aqkel__vfsx:
                        tpv__kbor = pbrh__hmlj
                    else:
                        tpv__kbor = op(lhs[nrji__yfiw], rhs[nrji__yfiw])
                    wfr__xst[nrji__yfiw] = tpv__kbor
                return wfr__xst
            return impl
        elif lhs == datetime_timedelta_array_type:

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(lhs)
                wfr__xst = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for nrji__yfiw in numba.parfors.parfor.internal_prange(n):
                    sjai__ottzg = bodo.libs.array_kernels.isna(lhs, nrji__yfiw)
                    if sjai__ottzg:
                        tpv__kbor = pbrh__hmlj
                    else:
                        tpv__kbor = op(lhs[nrji__yfiw], rhs)
                    wfr__xst[nrji__yfiw] = tpv__kbor
                return wfr__xst
            return impl
        elif rhs == datetime_timedelta_array_type:

            def impl(lhs, rhs):
                numba.parfors.parfor.init_prange()
                n = len(rhs)
                wfr__xst = bodo.libs.bool_arr_ext.alloc_bool_array(n)
                for nrji__yfiw in numba.parfors.parfor.internal_prange(n):
                    sjai__ottzg = bodo.libs.array_kernels.isna(rhs, nrji__yfiw)
                    if sjai__ottzg:
                        tpv__kbor = pbrh__hmlj
                    else:
                        tpv__kbor = op(lhs, rhs[nrji__yfiw])
                    wfr__xst[nrji__yfiw] = tpv__kbor
                return wfr__xst
            return impl
    return overload_date_arr_cmp


timedelta_unsupported_attrs = ['asm8', 'resolution_string', 'freq',
    'is_populated']
timedelta_unsupported_methods = ['isoformat']


def _intstall_pd_timedelta_unsupported():
    from bodo.utils.typing import create_unsupported_overload
    for bhavx__mhmnu in timedelta_unsupported_attrs:
        ocm__ffflo = 'pandas.Timedelta.' + bhavx__mhmnu
        overload_attribute(PDTimeDeltaType, bhavx__mhmnu)(
            create_unsupported_overload(ocm__ffflo))
    for tiqsd__gddfx in timedelta_unsupported_methods:
        ocm__ffflo = 'pandas.Timedelta.' + tiqsd__gddfx
        overload_method(PDTimeDeltaType, tiqsd__gddfx)(
            create_unsupported_overload(ocm__ffflo + '()'))


_intstall_pd_timedelta_unsupported()
