"""
Implement support for the various classes in pd.tseries.offsets.
"""
import operator
import llvmlite.binding as ll
import numpy as np
import pandas as pd
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, intrinsic, make_attribute_wrapper, models, overload, overload_attribute, overload_method, register_jitable, register_model, typeof_impl, unbox
from bodo.hiframes.datetime_date_ext import datetime_date_type
from bodo.hiframes.datetime_datetime_ext import datetime_datetime_type
from bodo.hiframes.pd_timestamp_ext import get_days_in_month, pd_timestamp_type
from bodo.libs import hdatetime_ext
from bodo.utils.typing import BodoError, create_unsupported_overload, is_overload_none
ll.add_symbol('box_date_offset', hdatetime_ext.box_date_offset)
ll.add_symbol('unbox_date_offset', hdatetime_ext.unbox_date_offset)


class MonthBeginType(types.Type):

    def __init__(self):
        super(MonthBeginType, self).__init__(name='MonthBeginType()')


month_begin_type = MonthBeginType()


@typeof_impl.register(pd.tseries.offsets.MonthBegin)
def typeof_month_begin(val, c):
    return month_begin_type


@register_model(MonthBeginType)
class MonthBeginModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pnhko__cluww = [('n', types.int64), ('normalize', types.boolean)]
        super(MonthBeginModel, self).__init__(dmm, fe_type, pnhko__cluww)


@box(MonthBeginType)
def box_month_begin(typ, val, c):
    qxsy__hvjqk = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    gdu__nqzdr = c.pyapi.long_from_longlong(qxsy__hvjqk.n)
    ernn__jzkl = c.pyapi.from_native_value(types.boolean, qxsy__hvjqk.
        normalize, c.env_manager)
    wgjk__qbtv = c.pyapi.unserialize(c.pyapi.serialize_object(pd.tseries.
        offsets.MonthBegin))
    ehu__ksgrb = c.pyapi.call_function_objargs(wgjk__qbtv, (gdu__nqzdr,
        ernn__jzkl))
    c.pyapi.decref(gdu__nqzdr)
    c.pyapi.decref(ernn__jzkl)
    c.pyapi.decref(wgjk__qbtv)
    return ehu__ksgrb


@unbox(MonthBeginType)
def unbox_month_begin(typ, val, c):
    gdu__nqzdr = c.pyapi.object_getattr_string(val, 'n')
    ernn__jzkl = c.pyapi.object_getattr_string(val, 'normalize')
    n = c.pyapi.long_as_longlong(gdu__nqzdr)
    normalize = c.pyapi.to_native_value(types.bool_, ernn__jzkl).value
    qxsy__hvjqk = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    qxsy__hvjqk.n = n
    qxsy__hvjqk.normalize = normalize
    c.pyapi.decref(gdu__nqzdr)
    c.pyapi.decref(ernn__jzkl)
    qofs__pcmag = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(qxsy__hvjqk._getvalue(), is_error=qofs__pcmag)


@overload(pd.tseries.offsets.MonthBegin, no_unliteral=True)
def MonthBegin(n=1, normalize=False):

    def impl(n=1, normalize=False):
        return init_month_begin(n, normalize)
    return impl


@intrinsic
def init_month_begin(typingctx, n, normalize):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        qxsy__hvjqk = cgutils.create_struct_proxy(typ)(context, builder)
        qxsy__hvjqk.n = args[0]
        qxsy__hvjqk.normalize = args[1]
        return qxsy__hvjqk._getvalue()
    return MonthBeginType()(n, normalize), codegen


make_attribute_wrapper(MonthBeginType, 'n', 'n')
make_attribute_wrapper(MonthBeginType, 'normalize', 'normalize')


@register_jitable
def calculate_month_begin_date(year, month, day, n):
    if n <= 0:
        if day > 1:
            n += 1
    month = month + n
    month -= 1
    year += month // 12
    month = month % 12 + 1
    day = 1
    return year, month, day


def overload_add_operator_month_begin_offset_type(lhs, rhs):
    if lhs == month_begin_type and rhs == datetime_datetime_type:

        def impl(lhs, rhs):
            year, month, day = calculate_month_begin_date(rhs.year, rhs.
                month, rhs.day, lhs.n)
            if lhs.normalize:
                return pd.Timestamp(year=year, month=month, day=day)
            else:
                return pd.Timestamp(year=year, month=month, day=day, hour=
                    rhs.hour, minute=rhs.minute, second=rhs.second,
                    microsecond=rhs.microsecond)
        return impl
    if lhs == month_begin_type and rhs == pd_timestamp_type:

        def impl(lhs, rhs):
            year, month, day = calculate_month_begin_date(rhs.year, rhs.
                month, rhs.day, lhs.n)
            if lhs.normalize:
                return pd.Timestamp(year=year, month=month, day=day)
            else:
                return pd.Timestamp(year=year, month=month, day=day, hour=
                    rhs.hour, minute=rhs.minute, second=rhs.second,
                    microsecond=rhs.microsecond, nanosecond=rhs.nanosecond)
        return impl
    if lhs == month_begin_type and rhs == datetime_date_type:

        def impl(lhs, rhs):
            year, month, day = calculate_month_begin_date(rhs.year, rhs.
                month, rhs.day, lhs.n)
            return pd.Timestamp(year=year, month=month, day=day)
        return impl
    if lhs in [datetime_datetime_type, pd_timestamp_type, datetime_date_type
        ] and rhs == month_begin_type:

        def impl(lhs, rhs):
            return rhs + lhs
        return impl
    raise BodoError(
        f'add operator not supported for data types {lhs} and {rhs}.')


class MonthEndType(types.Type):

    def __init__(self):
        super(MonthEndType, self).__init__(name='MonthEndType()')


month_end_type = MonthEndType()


@typeof_impl.register(pd.tseries.offsets.MonthEnd)
def typeof_month_end(val, c):
    return month_end_type


@register_model(MonthEndType)
class MonthEndModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pnhko__cluww = [('n', types.int64), ('normalize', types.boolean)]
        super(MonthEndModel, self).__init__(dmm, fe_type, pnhko__cluww)


@box(MonthEndType)
def box_month_end(typ, val, c):
    eat__azmgz = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    gdu__nqzdr = c.pyapi.long_from_longlong(eat__azmgz.n)
    ernn__jzkl = c.pyapi.from_native_value(types.boolean, eat__azmgz.
        normalize, c.env_manager)
    ttkf__qakd = c.pyapi.unserialize(c.pyapi.serialize_object(pd.tseries.
        offsets.MonthEnd))
    ehu__ksgrb = c.pyapi.call_function_objargs(ttkf__qakd, (gdu__nqzdr,
        ernn__jzkl))
    c.pyapi.decref(gdu__nqzdr)
    c.pyapi.decref(ernn__jzkl)
    c.pyapi.decref(ttkf__qakd)
    return ehu__ksgrb


@unbox(MonthEndType)
def unbox_month_end(typ, val, c):
    gdu__nqzdr = c.pyapi.object_getattr_string(val, 'n')
    ernn__jzkl = c.pyapi.object_getattr_string(val, 'normalize')
    n = c.pyapi.long_as_longlong(gdu__nqzdr)
    normalize = c.pyapi.to_native_value(types.bool_, ernn__jzkl).value
    eat__azmgz = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    eat__azmgz.n = n
    eat__azmgz.normalize = normalize
    c.pyapi.decref(gdu__nqzdr)
    c.pyapi.decref(ernn__jzkl)
    qofs__pcmag = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(eat__azmgz._getvalue(), is_error=qofs__pcmag)


@overload(pd.tseries.offsets.MonthEnd, no_unliteral=True)
def MonthEnd(n=1, normalize=False):

    def impl(n=1, normalize=False):
        return init_month_end(n, normalize)
    return impl


@intrinsic
def init_month_end(typingctx, n, normalize):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        eat__azmgz = cgutils.create_struct_proxy(typ)(context, builder)
        eat__azmgz.n = args[0]
        eat__azmgz.normalize = args[1]
        return eat__azmgz._getvalue()
    return MonthEndType()(n, normalize), codegen


make_attribute_wrapper(MonthEndType, 'n', 'n')
make_attribute_wrapper(MonthEndType, 'normalize', 'normalize')


@lower_constant(MonthBeginType)
@lower_constant(MonthEndType)
def lower_constant_month_end(context, builder, ty, pyval):
    n = context.get_constant(types.int64, pyval.n)
    normalize = context.get_constant(types.boolean, pyval.normalize)
    return lir.Constant.literal_struct([n, normalize])


@register_jitable
def calculate_month_end_date(year, month, day, n):
    if n > 0:
        eat__azmgz = get_days_in_month(year, month)
        if eat__azmgz > day:
            n -= 1
    month = month + n
    month -= 1
    year += month // 12
    month = month % 12 + 1
    day = get_days_in_month(year, month)
    return year, month, day


def overload_add_operator_month_end_offset_type(lhs, rhs):
    if lhs == month_end_type and rhs == datetime_datetime_type:

        def impl(lhs, rhs):
            year, month, day = calculate_month_end_date(rhs.year, rhs.month,
                rhs.day, lhs.n)
            if lhs.normalize:
                return pd.Timestamp(year=year, month=month, day=day)
            else:
                return pd.Timestamp(year=year, month=month, day=day, hour=
                    rhs.hour, minute=rhs.minute, second=rhs.second,
                    microsecond=rhs.microsecond)
        return impl
    if lhs == month_end_type and rhs == pd_timestamp_type:

        def impl(lhs, rhs):
            year, month, day = calculate_month_end_date(rhs.year, rhs.month,
                rhs.day, lhs.n)
            if lhs.normalize:
                return pd.Timestamp(year=year, month=month, day=day)
            else:
                return pd.Timestamp(year=year, month=month, day=day, hour=
                    rhs.hour, minute=rhs.minute, second=rhs.second,
                    microsecond=rhs.microsecond, nanosecond=rhs.nanosecond)
        return impl
    if lhs == month_end_type and rhs == datetime_date_type:

        def impl(lhs, rhs):
            year, month, day = calculate_month_end_date(rhs.year, rhs.month,
                rhs.day, lhs.n)
            return pd.Timestamp(year=year, month=month, day=day)
        return impl
    if lhs in [datetime_datetime_type, pd_timestamp_type, datetime_date_type
        ] and rhs == month_end_type:

        def impl(lhs, rhs):
            return rhs + lhs
        return impl
    raise BodoError(
        f'add operator not supported for data types {lhs} and {rhs}.')


def overload_mul_date_offset_types(lhs, rhs):
    if lhs == month_begin_type:

        def impl(lhs, rhs):
            return pd.tseries.offsets.MonthBegin(lhs.n * rhs, lhs.normalize)
    if lhs == month_end_type:

        def impl(lhs, rhs):
            return pd.tseries.offsets.MonthEnd(lhs.n * rhs, lhs.normalize)
    if lhs == week_type:

        def impl(lhs, rhs):
            return pd.tseries.offsets.Week(lhs.n * rhs, lhs.normalize, lhs.
                weekday)
    if lhs == date_offset_type:

        def impl(lhs, rhs):
            n = lhs.n * rhs
            normalize = lhs.normalize
            nanoseconds = lhs._nanoseconds
            nanosecond = lhs._nanosecond
            if lhs._has_kws:
                years = lhs._years
                months = lhs._months
                weeks = lhs._weeks
                days = lhs._days
                hours = lhs._hours
                minutes = lhs._minutes
                seconds = lhs._seconds
                microseconds = lhs._microseconds
                year = lhs._year
                month = lhs._month
                day = lhs._day
                weekday = lhs._weekday
                hour = lhs._hour
                minute = lhs._minute
                second = lhs._second
                microsecond = lhs._microsecond
                return pd.tseries.offsets.DateOffset(n, normalize, years,
                    months, weeks, days, hours, minutes, seconds,
                    microseconds, nanoseconds, year, month, day, weekday,
                    hour, minute, second, microsecond, nanosecond)
            else:
                return pd.tseries.offsets.DateOffset(n, normalize,
                    nanoseconds=nanoseconds, nanosecond=nanosecond)
    if rhs in [week_type, month_end_type, month_begin_type, date_offset_type]:

        def impl(lhs, rhs):
            return rhs * lhs
        return impl
    return impl


class DateOffsetType(types.Type):

    def __init__(self):
        super(DateOffsetType, self).__init__(name='DateOffsetType()')


date_offset_type = DateOffsetType()
date_offset_fields = ['years', 'months', 'weeks', 'days', 'hours',
    'minutes', 'seconds', 'microseconds', 'nanoseconds', 'year', 'month',
    'day', 'weekday', 'hour', 'minute', 'second', 'microsecond', 'nanosecond']


@typeof_impl.register(pd.tseries.offsets.DateOffset)
def type_of_date_offset(val, c):
    return date_offset_type


@register_model(DateOffsetType)
class DateOffsetModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pnhko__cluww = [('n', types.int64), ('normalize', types.boolean), (
            'years', types.int64), ('months', types.int64), ('weeks', types
            .int64), ('days', types.int64), ('hours', types.int64), (
            'minutes', types.int64), ('seconds', types.int64), (
            'microseconds', types.int64), ('nanoseconds', types.int64), (
            'year', types.int64), ('month', types.int64), ('day', types.
            int64), ('weekday', types.int64), ('hour', types.int64), (
            'minute', types.int64), ('second', types.int64), ('microsecond',
            types.int64), ('nanosecond', types.int64), ('has_kws', types.
            boolean)]
        super(DateOffsetModel, self).__init__(dmm, fe_type, pnhko__cluww)


@box(DateOffsetType)
def box_date_offset(typ, val, c):
    uqrbm__voxt = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    eeg__xfdn = c.builder.alloca(lir.IntType(64), size=lir.Constant(lir.
        IntType(64), 18))
    for nvbh__rngbw, oqqjo__nsbfv in enumerate(date_offset_fields):
        c.builder.store(getattr(uqrbm__voxt, oqqjo__nsbfv), c.builder.
            inttoptr(c.builder.add(c.builder.ptrtoint(eeg__xfdn, lir.
            IntType(64)), lir.Constant(lir.IntType(64), 8 * nvbh__rngbw)),
            lir.IntType(64).as_pointer()))
    vlsqf__povti = lir.FunctionType(c.pyapi.pyobj, [lir.IntType(64), lir.
        IntType(1), lir.IntType(64).as_pointer(), lir.IntType(1)])
    ubiqn__itq = cgutils.get_or_insert_function(c.builder.module,
        vlsqf__povti, name='box_date_offset')
    xrlqr__meijb = c.builder.call(ubiqn__itq, [uqrbm__voxt.n, uqrbm__voxt.
        normalize, eeg__xfdn, uqrbm__voxt.has_kws])
    c.context.nrt.decref(c.builder, typ, val)
    return xrlqr__meijb


@unbox(DateOffsetType)
def unbox_date_offset(typ, val, c):
    gdu__nqzdr = c.pyapi.object_getattr_string(val, 'n')
    ernn__jzkl = c.pyapi.object_getattr_string(val, 'normalize')
    n = c.pyapi.long_as_longlong(gdu__nqzdr)
    normalize = c.pyapi.to_native_value(types.bool_, ernn__jzkl).value
    eeg__xfdn = c.builder.alloca(lir.IntType(64), size=lir.Constant(lir.
        IntType(64), 18))
    vlsqf__povti = lir.FunctionType(lir.IntType(1), [lir.IntType(8).
        as_pointer(), lir.IntType(64).as_pointer()])
    nbhb__kpyux = cgutils.get_or_insert_function(c.builder.module,
        vlsqf__povti, name='unbox_date_offset')
    has_kws = c.builder.call(nbhb__kpyux, [val, eeg__xfdn])
    uqrbm__voxt = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    uqrbm__voxt.n = n
    uqrbm__voxt.normalize = normalize
    for nvbh__rngbw, oqqjo__nsbfv in enumerate(date_offset_fields):
        setattr(uqrbm__voxt, oqqjo__nsbfv, c.builder.load(c.builder.
            inttoptr(c.builder.add(c.builder.ptrtoint(eeg__xfdn, lir.
            IntType(64)), lir.Constant(lir.IntType(64), 8 * nvbh__rngbw)),
            lir.IntType(64).as_pointer())))
    uqrbm__voxt.has_kws = has_kws
    c.pyapi.decref(gdu__nqzdr)
    c.pyapi.decref(ernn__jzkl)
    qofs__pcmag = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(uqrbm__voxt._getvalue(), is_error=qofs__pcmag)


@lower_constant(DateOffsetType)
def lower_constant_date_offset(context, builder, ty, pyval):
    n = context.get_constant(types.int64, pyval.n)
    normalize = context.get_constant(types.boolean, pyval.normalize)
    fnjv__umpt = [n, normalize]
    has_kws = False
    fzgia__fko = [0] * 9 + [-1] * 9
    for nvbh__rngbw, oqqjo__nsbfv in enumerate(date_offset_fields):
        if hasattr(pyval, oqqjo__nsbfv):
            qfwj__rzzwz = context.get_constant(types.int64, getattr(pyval,
                oqqjo__nsbfv))
            if oqqjo__nsbfv != 'nanoseconds' and oqqjo__nsbfv != 'nanosecond':
                has_kws = True
        else:
            qfwj__rzzwz = context.get_constant(types.int64, fzgia__fko[
                nvbh__rngbw])
        fnjv__umpt.append(qfwj__rzzwz)
    has_kws = context.get_constant(types.boolean, has_kws)
    fnjv__umpt.append(has_kws)
    return lir.Constant.literal_struct(fnjv__umpt)


@overload(pd.tseries.offsets.DateOffset, no_unliteral=True)
def DateOffset(n=1, normalize=False, years=None, months=None, weeks=None,
    days=None, hours=None, minutes=None, seconds=None, microseconds=None,
    nanoseconds=None, year=None, month=None, day=None, weekday=None, hour=
    None, minute=None, second=None, microsecond=None, nanosecond=None):
    has_kws = False
    nufm__abbd = [years, months, weeks, days, hours, minutes, seconds,
        microseconds, year, month, day, weekday, hour, minute, second,
        microsecond]
    for oqy__kvt in nufm__abbd:
        if not is_overload_none(oqy__kvt):
            has_kws = True
            break

    def impl(n=1, normalize=False, years=None, months=None, weeks=None,
        days=None, hours=None, minutes=None, seconds=None, microseconds=
        None, nanoseconds=None, year=None, month=None, day=None, weekday=
        None, hour=None, minute=None, second=None, microsecond=None,
        nanosecond=None):
        years = 0 if years is None else years
        months = 0 if months is None else months
        weeks = 0 if weeks is None else weeks
        days = 0 if days is None else days
        hours = 0 if hours is None else hours
        minutes = 0 if minutes is None else minutes
        seconds = 0 if seconds is None else seconds
        microseconds = 0 if microseconds is None else microseconds
        nanoseconds = 0 if nanoseconds is None else nanoseconds
        year = -1 if year is None else year
        month = -1 if month is None else month
        weekday = -1 if weekday is None else weekday
        day = -1 if day is None else day
        hour = -1 if hour is None else hour
        minute = -1 if minute is None else minute
        second = -1 if second is None else second
        microsecond = -1 if microsecond is None else microsecond
        nanosecond = -1 if nanosecond is None else nanosecond
        return init_date_offset(n, normalize, years, months, weeks, days,
            hours, minutes, seconds, microseconds, nanoseconds, year, month,
            day, weekday, hour, minute, second, microsecond, nanosecond,
            has_kws)
    return impl


@intrinsic
def init_date_offset(typingctx, n, normalize, years, months, weeks, days,
    hours, minutes, seconds, microseconds, nanoseconds, year, month, day,
    weekday, hour, minute, second, microsecond, nanosecond, has_kws):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        uqrbm__voxt = cgutils.create_struct_proxy(typ)(context, builder)
        uqrbm__voxt.n = args[0]
        uqrbm__voxt.normalize = args[1]
        uqrbm__voxt.years = args[2]
        uqrbm__voxt.months = args[3]
        uqrbm__voxt.weeks = args[4]
        uqrbm__voxt.days = args[5]
        uqrbm__voxt.hours = args[6]
        uqrbm__voxt.minutes = args[7]
        uqrbm__voxt.seconds = args[8]
        uqrbm__voxt.microseconds = args[9]
        uqrbm__voxt.nanoseconds = args[10]
        uqrbm__voxt.year = args[11]
        uqrbm__voxt.month = args[12]
        uqrbm__voxt.day = args[13]
        uqrbm__voxt.weekday = args[14]
        uqrbm__voxt.hour = args[15]
        uqrbm__voxt.minute = args[16]
        uqrbm__voxt.second = args[17]
        uqrbm__voxt.microsecond = args[18]
        uqrbm__voxt.nanosecond = args[19]
        uqrbm__voxt.has_kws = args[20]
        return uqrbm__voxt._getvalue()
    return DateOffsetType()(n, normalize, years, months, weeks, days, hours,
        minutes, seconds, microseconds, nanoseconds, year, month, day,
        weekday, hour, minute, second, microsecond, nanosecond, has_kws
        ), codegen


make_attribute_wrapper(DateOffsetType, 'n', 'n')
make_attribute_wrapper(DateOffsetType, 'normalize', 'normalize')
make_attribute_wrapper(DateOffsetType, 'years', '_years')
make_attribute_wrapper(DateOffsetType, 'months', '_months')
make_attribute_wrapper(DateOffsetType, 'weeks', '_weeks')
make_attribute_wrapper(DateOffsetType, 'days', '_days')
make_attribute_wrapper(DateOffsetType, 'hours', '_hours')
make_attribute_wrapper(DateOffsetType, 'minutes', '_minutes')
make_attribute_wrapper(DateOffsetType, 'seconds', '_seconds')
make_attribute_wrapper(DateOffsetType, 'microseconds', '_microseconds')
make_attribute_wrapper(DateOffsetType, 'nanoseconds', '_nanoseconds')
make_attribute_wrapper(DateOffsetType, 'year', '_year')
make_attribute_wrapper(DateOffsetType, 'month', '_month')
make_attribute_wrapper(DateOffsetType, 'weekday', '_weekday')
make_attribute_wrapper(DateOffsetType, 'day', '_day')
make_attribute_wrapper(DateOffsetType, 'hour', '_hour')
make_attribute_wrapper(DateOffsetType, 'minute', '_minute')
make_attribute_wrapper(DateOffsetType, 'second', '_second')
make_attribute_wrapper(DateOffsetType, 'microsecond', '_microsecond')
make_attribute_wrapper(DateOffsetType, 'nanosecond', '_nanosecond')
make_attribute_wrapper(DateOffsetType, 'has_kws', '_has_kws')


@register_jitable
def relative_delta_addition(dateoffset, ts):
    if dateoffset._has_kws:
        euio__owm = -1 if dateoffset.n < 0 else 1
        for wtdu__zee in range(np.abs(dateoffset.n)):
            year = ts.year
            month = ts.month
            day = ts.day
            hour = ts.hour
            minute = ts.minute
            second = ts.second
            microsecond = ts.microsecond
            nanosecond = ts.nanosecond
            if dateoffset._year != -1:
                year = dateoffset._year
            year += euio__owm * dateoffset._years
            if dateoffset._month != -1:
                month = dateoffset._month
            month += euio__owm * dateoffset._months
            year, month, fdqq__ihto = calculate_month_end_date(year, month,
                day, 0)
            if day > fdqq__ihto:
                day = fdqq__ihto
            if dateoffset._day != -1:
                day = dateoffset._day
            if dateoffset._hour != -1:
                hour = dateoffset._hour
            if dateoffset._minute != -1:
                minute = dateoffset._minute
            if dateoffset._second != -1:
                second = dateoffset._second
            if dateoffset._microsecond != -1:
                microsecond = dateoffset._microsecond
            ts = pd.Timestamp(year=year, month=month, day=day, hour=hour,
                minute=minute, second=second, microsecond=microsecond,
                nanosecond=nanosecond)
            qvgd__bdkl = pd.Timedelta(days=dateoffset._days + 7 *
                dateoffset._weeks, hours=dateoffset._hours, minutes=
                dateoffset._minutes, seconds=dateoffset._seconds,
                microseconds=dateoffset._microseconds)
            if euio__owm == -1:
                qvgd__bdkl = -qvgd__bdkl
            ts = ts + qvgd__bdkl
            if dateoffset._weekday != -1:
                goemw__yse = ts.weekday()
                ujp__irug = (dateoffset._weekday - goemw__yse) % 7
                ts = ts + pd.Timedelta(days=ujp__irug)
        return ts
    else:
        return pd.Timedelta(days=dateoffset.n) + ts


def overload_add_operator_date_offset_type(lhs, rhs):
    if lhs == date_offset_type and rhs == pd_timestamp_type:

        def impl(lhs, rhs):
            ts = relative_delta_addition(lhs, rhs)
            if lhs.normalize:
                return ts.normalize()
            return ts
        return impl
    if lhs == date_offset_type and rhs in [datetime_date_type,
        datetime_datetime_type]:

        def impl(lhs, rhs):
            ts = relative_delta_addition(lhs, pd.Timestamp(rhs))
            if lhs.normalize:
                return ts.normalize()
            return ts
        return impl
    if lhs in [datetime_datetime_type, pd_timestamp_type, datetime_date_type
        ] and rhs == date_offset_type:

        def impl(lhs, rhs):
            return rhs + lhs
        return impl
    raise BodoError(
        f'add operator not supported for data types {lhs} and {rhs}.')


def overload_sub_operator_offsets(lhs, rhs):
    if lhs in [datetime_datetime_type, pd_timestamp_type, datetime_date_type
        ] and rhs in [date_offset_type, month_begin_type, month_end_type,
        week_type]:

        def impl(lhs, rhs):
            return lhs + -rhs
        return impl


@overload(operator.neg, no_unliteral=True)
def overload_neg(lhs):
    if lhs == month_begin_type:

        def impl(lhs):
            return pd.tseries.offsets.MonthBegin(-lhs.n, lhs.normalize)
    elif lhs == month_end_type:

        def impl(lhs):
            return pd.tseries.offsets.MonthEnd(-lhs.n, lhs.normalize)
    elif lhs == week_type:

        def impl(lhs):
            return pd.tseries.offsets.Week(-lhs.n, lhs.normalize, lhs.weekday)
    elif lhs == date_offset_type:

        def impl(lhs):
            n = -lhs.n
            normalize = lhs.normalize
            nanoseconds = lhs._nanoseconds
            nanosecond = lhs._nanosecond
            if lhs._has_kws:
                years = lhs._years
                months = lhs._months
                weeks = lhs._weeks
                days = lhs._days
                hours = lhs._hours
                minutes = lhs._minutes
                seconds = lhs._seconds
                microseconds = lhs._microseconds
                year = lhs._year
                month = lhs._month
                day = lhs._day
                weekday = lhs._weekday
                hour = lhs._hour
                minute = lhs._minute
                second = lhs._second
                microsecond = lhs._microsecond
                return pd.tseries.offsets.DateOffset(n, normalize, years,
                    months, weeks, days, hours, minutes, seconds,
                    microseconds, nanoseconds, year, month, day, weekday,
                    hour, minute, second, microsecond, nanosecond)
            else:
                return pd.tseries.offsets.DateOffset(n, normalize,
                    nanoseconds=nanoseconds, nanosecond=nanosecond)
    else:
        return
    return impl


def is_offsets_type(val):
    return val in [date_offset_type, month_begin_type, month_end_type,
        week_type]


class WeekType(types.Type):

    def __init__(self):
        super(WeekType, self).__init__(name='WeekType()')


week_type = WeekType()


@typeof_impl.register(pd.tseries.offsets.Week)
def typeof_week(val, c):
    return week_type


@register_model(WeekType)
class WeekModel(models.StructModel):

    def __init__(self, dmm, fe_type):
        pnhko__cluww = [('n', types.int64), ('normalize', types.boolean), (
            'weekday', types.int64)]
        super(WeekModel, self).__init__(dmm, fe_type, pnhko__cluww)


make_attribute_wrapper(WeekType, 'n', 'n')
make_attribute_wrapper(WeekType, 'normalize', 'normalize')
make_attribute_wrapper(WeekType, 'weekday', 'weekday')


@overload(pd.tseries.offsets.Week, no_unliteral=True)
def Week(n=1, normalize=False, weekday=None):

    def impl(n=1, normalize=False, weekday=None):
        gfjqb__vuzx = -1 if weekday is None else weekday
        return init_week(n, normalize, gfjqb__vuzx)
    return impl


@intrinsic
def init_week(typingctx, n, normalize, weekday):

    def codegen(context, builder, signature, args):
        typ = signature.return_type
        jkmc__fqa = cgutils.create_struct_proxy(typ)(context, builder)
        jkmc__fqa.n = args[0]
        jkmc__fqa.normalize = args[1]
        jkmc__fqa.weekday = args[2]
        return jkmc__fqa._getvalue()
    return WeekType()(n, normalize, weekday), codegen


@lower_constant(WeekType)
def lower_constant_week(context, builder, ty, pyval):
    n = context.get_constant(types.int64, pyval.n)
    normalize = context.get_constant(types.boolean, pyval.normalize)
    if pyval.weekday is not None:
        weekday = context.get_constant(types.int64, pyval.weekday)
    else:
        weekday = context.get_constant(types.int64, -1)
    return lir.Constant.literal_struct([n, normalize, weekday])


@box(WeekType)
def box_week(typ, val, c):
    jkmc__fqa = cgutils.create_struct_proxy(typ)(c.context, c.builder,
        value=val)
    gdu__nqzdr = c.pyapi.long_from_longlong(jkmc__fqa.n)
    ernn__jzkl = c.pyapi.from_native_value(types.boolean, jkmc__fqa.
        normalize, c.env_manager)
    eybwx__vzud = c.pyapi.long_from_longlong(jkmc__fqa.weekday)
    ngq__amp = c.pyapi.unserialize(c.pyapi.serialize_object(pd.tseries.
        offsets.Week))
    eue__vsdgk = c.builder.icmp_signed('!=', lir.Constant(lir.IntType(64), 
        -1), jkmc__fqa.weekday)
    with c.builder.if_else(eue__vsdgk) as (btuw__pwq, hda__ocnpf):
        with btuw__pwq:
            kupa__bem = c.pyapi.call_function_objargs(ngq__amp, (gdu__nqzdr,
                ernn__jzkl, eybwx__vzud))
            lxk__ffxjv = c.builder.block
        with hda__ocnpf:
            xuiq__iawcl = c.pyapi.call_function_objargs(ngq__amp, (
                gdu__nqzdr, ernn__jzkl))
            guv__ziqk = c.builder.block
    ehu__ksgrb = c.builder.phi(kupa__bem.type)
    ehu__ksgrb.add_incoming(kupa__bem, lxk__ffxjv)
    ehu__ksgrb.add_incoming(xuiq__iawcl, guv__ziqk)
    c.pyapi.decref(eybwx__vzud)
    c.pyapi.decref(gdu__nqzdr)
    c.pyapi.decref(ernn__jzkl)
    c.pyapi.decref(ngq__amp)
    return ehu__ksgrb


@unbox(WeekType)
def unbox_week(typ, val, c):
    gdu__nqzdr = c.pyapi.object_getattr_string(val, 'n')
    ernn__jzkl = c.pyapi.object_getattr_string(val, 'normalize')
    eybwx__vzud = c.pyapi.object_getattr_string(val, 'weekday')
    n = c.pyapi.long_as_longlong(gdu__nqzdr)
    normalize = c.pyapi.to_native_value(types.bool_, ernn__jzkl).value
    fpypy__fkbye = c.pyapi.make_none()
    vata__czwph = c.builder.icmp_unsigned('==', eybwx__vzud, fpypy__fkbye)
    with c.builder.if_else(vata__czwph) as (hda__ocnpf, btuw__pwq):
        with btuw__pwq:
            kupa__bem = c.pyapi.long_as_longlong(eybwx__vzud)
            lxk__ffxjv = c.builder.block
        with hda__ocnpf:
            xuiq__iawcl = lir.Constant(lir.IntType(64), -1)
            guv__ziqk = c.builder.block
    ehu__ksgrb = c.builder.phi(kupa__bem.type)
    ehu__ksgrb.add_incoming(kupa__bem, lxk__ffxjv)
    ehu__ksgrb.add_incoming(xuiq__iawcl, guv__ziqk)
    jkmc__fqa = cgutils.create_struct_proxy(typ)(c.context, c.builder)
    jkmc__fqa.n = n
    jkmc__fqa.normalize = normalize
    jkmc__fqa.weekday = ehu__ksgrb
    c.pyapi.decref(gdu__nqzdr)
    c.pyapi.decref(ernn__jzkl)
    c.pyapi.decref(eybwx__vzud)
    qofs__pcmag = cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return NativeValue(jkmc__fqa._getvalue(), is_error=qofs__pcmag)


def overload_add_operator_week_offset_type(lhs, rhs):
    if lhs == week_type and rhs == pd_timestamp_type:

        def impl(lhs, rhs):
            utnf__okoki = calculate_week_date(lhs.n, lhs.weekday, rhs.weekday()
                )
            if lhs.normalize:
                blfxd__bmf = pd.Timestamp(year=rhs.year, month=rhs.month,
                    day=rhs.day)
            else:
                blfxd__bmf = rhs
            return blfxd__bmf + utnf__okoki
        return impl
    if lhs == week_type and rhs == datetime_datetime_type:

        def impl(lhs, rhs):
            utnf__okoki = calculate_week_date(lhs.n, lhs.weekday, rhs.weekday()
                )
            if lhs.normalize:
                blfxd__bmf = pd.Timestamp(year=rhs.year, month=rhs.month,
                    day=rhs.day)
            else:
                blfxd__bmf = pd.Timestamp(year=rhs.year, month=rhs.month,
                    day=rhs.day, hour=rhs.hour, minute=rhs.minute, second=
                    rhs.second, microsecond=rhs.microsecond)
            return blfxd__bmf + utnf__okoki
        return impl
    if lhs == week_type and rhs == datetime_date_type:

        def impl(lhs, rhs):
            utnf__okoki = calculate_week_date(lhs.n, lhs.weekday, rhs.weekday()
                )
            return rhs + utnf__okoki
        return impl
    if lhs in [datetime_datetime_type, pd_timestamp_type, datetime_date_type
        ] and rhs == week_type:

        def impl(lhs, rhs):
            return rhs + lhs
        return impl
    raise BodoError(
        f'add operator not supported for data types {lhs} and {rhs}.')


@register_jitable
def calculate_week_date(n, weekday, other_weekday):
    if weekday == -1:
        return pd.Timedelta(weeks=n)
    if weekday != other_weekday:
        kekr__nqmtf = (weekday - other_weekday) % 7
        if n > 0:
            n = n - 1
    return pd.Timedelta(weeks=n, days=kekr__nqmtf)


date_offset_unsupported_attrs = {'base', 'freqstr', 'kwds', 'name', 'nanos',
    'rule_code'}
date_offset_unsupported = {'__call__', 'rollback', 'rollforward',
    'is_month_start', 'is_month_end', 'apply', 'apply_index', 'copy',
    'isAnchored', 'onOffset', 'is_anchored', 'is_on_offset',
    'is_quarter_start', 'is_quarter_end', 'is_year_start', 'is_year_end'}
month_end_unsupported_attrs = {'base', 'freqstr', 'kwds', 'name', 'nanos',
    'rule_code'}
month_end_unsupported = {'__call__', 'rollback', 'rollforward', 'apply',
    'apply_index', 'copy', 'isAnchored', 'onOffset', 'is_anchored',
    'is_on_offset', 'is_month_start', 'is_month_end', 'is_quarter_start',
    'is_quarter_end', 'is_year_start', 'is_year_end'}
month_begin_unsupported_attrs = {'basefreqstr', 'kwds', 'name', 'nanos',
    'rule_code'}
month_begin_unsupported = {'__call__', 'rollback', 'rollforward', 'apply',
    'apply_index', 'copy', 'isAnchored', 'onOffset', 'is_anchored',
    'is_on_offset', 'is_month_start', 'is_month_end', 'is_quarter_start',
    'is_quarter_end', 'is_year_start', 'is_year_end'}
week_unsupported_attrs = {'basefreqstr', 'kwds', 'name', 'nanos', 'rule_code'}
week_unsupported = {'__call__', 'rollback', 'rollforward', 'apply',
    'apply_index', 'copy', 'isAnchored', 'onOffset', 'is_anchored',
    'is_on_offset', 'is_month_start', 'is_month_end', 'is_quarter_start',
    'is_quarter_end', 'is_year_start', 'is_year_end'}
offsets_unsupported = {pd.tseries.offsets.BusinessDay, pd.tseries.offsets.
    BDay, pd.tseries.offsets.BusinessHour, pd.tseries.offsets.
    CustomBusinessDay, pd.tseries.offsets.CDay, pd.tseries.offsets.
    CustomBusinessHour, pd.tseries.offsets.BusinessMonthEnd, pd.tseries.
    offsets.BMonthEnd, pd.tseries.offsets.BusinessMonthBegin, pd.tseries.
    offsets.BMonthBegin, pd.tseries.offsets.CustomBusinessMonthEnd, pd.
    tseries.offsets.CBMonthEnd, pd.tseries.offsets.CustomBusinessMonthBegin,
    pd.tseries.offsets.CBMonthBegin, pd.tseries.offsets.SemiMonthEnd, pd.
    tseries.offsets.SemiMonthBegin, pd.tseries.offsets.WeekOfMonth, pd.
    tseries.offsets.LastWeekOfMonth, pd.tseries.offsets.BQuarterEnd, pd.
    tseries.offsets.BQuarterBegin, pd.tseries.offsets.QuarterEnd, pd.
    tseries.offsets.QuarterBegin, pd.tseries.offsets.BYearEnd, pd.tseries.
    offsets.BYearBegin, pd.tseries.offsets.YearEnd, pd.tseries.offsets.
    YearBegin, pd.tseries.offsets.FY5253, pd.tseries.offsets.FY5253Quarter,
    pd.tseries.offsets.Easter, pd.tseries.offsets.Tick, pd.tseries.offsets.
    Day, pd.tseries.offsets.Hour, pd.tseries.offsets.Minute, pd.tseries.
    offsets.Second, pd.tseries.offsets.Milli, pd.tseries.offsets.Micro, pd.
    tseries.offsets.Nano}
frequencies_unsupported = {pd.tseries.frequencies.to_offset}


def _install_date_offsets_unsupported():
    for kzg__vzty in date_offset_unsupported_attrs:
        tyowe__aiate = 'pandas.tseries.offsets.DateOffset.' + kzg__vzty
        overload_attribute(DateOffsetType, kzg__vzty)(
            create_unsupported_overload(tyowe__aiate))
    for kzg__vzty in date_offset_unsupported:
        tyowe__aiate = 'pandas.tseries.offsets.DateOffset.' + kzg__vzty
        overload_method(DateOffsetType, kzg__vzty)(create_unsupported_overload
            (tyowe__aiate))


def _install_month_begin_unsupported():
    for kzg__vzty in month_begin_unsupported_attrs:
        tyowe__aiate = 'pandas.tseries.offsets.MonthBegin.' + kzg__vzty
        overload_attribute(MonthBeginType, kzg__vzty)(
            create_unsupported_overload(tyowe__aiate))
    for kzg__vzty in month_begin_unsupported:
        tyowe__aiate = 'pandas.tseries.offsets.MonthBegin.' + kzg__vzty
        overload_method(MonthBeginType, kzg__vzty)(create_unsupported_overload
            (tyowe__aiate))


def _install_month_end_unsupported():
    for kzg__vzty in date_offset_unsupported_attrs:
        tyowe__aiate = 'pandas.tseries.offsets.MonthEnd.' + kzg__vzty
        overload_attribute(MonthEndType, kzg__vzty)(create_unsupported_overload
            (tyowe__aiate))
    for kzg__vzty in date_offset_unsupported:
        tyowe__aiate = 'pandas.tseries.offsets.MonthEnd.' + kzg__vzty
        overload_method(MonthEndType, kzg__vzty)(create_unsupported_overload
            (tyowe__aiate))


def _install_week_unsupported():
    for kzg__vzty in week_unsupported_attrs:
        tyowe__aiate = 'pandas.tseries.offsets.Week.' + kzg__vzty
        overload_attribute(WeekType, kzg__vzty)(create_unsupported_overload
            (tyowe__aiate))
    for kzg__vzty in week_unsupported:
        tyowe__aiate = 'pandas.tseries.offsets.Week.' + kzg__vzty
        overload_method(WeekType, kzg__vzty)(create_unsupported_overload(
            tyowe__aiate))


def _install_offsets_unsupported():
    for qfwj__rzzwz in offsets_unsupported:
        tyowe__aiate = 'pandas.tseries.offsets.' + qfwj__rzzwz.__name__
        overload(qfwj__rzzwz)(create_unsupported_overload(tyowe__aiate))


def _install_frequencies_unsupported():
    for qfwj__rzzwz in frequencies_unsupported:
        tyowe__aiate = 'pandas.tseries.frequencies.' + qfwj__rzzwz.__name__
        overload(qfwj__rzzwz)(create_unsupported_overload(tyowe__aiate))


_install_date_offsets_unsupported()
_install_month_begin_unsupported()
_install_month_end_unsupported()
_install_week_unsupported()
_install_offsets_unsupported()
_install_frequencies_unsupported()
